from collections import OrderedDict

from hkube_python_wrapper.communication.zmq.streaming.producer.Flow import Flow


class MessageQueue(object):
    def __init__(self, consumerTypes, me):
        self.me = me
        self.consumerTypes = consumerTypes

        self.indexPerConsumer = OrderedDict()
        self.sent = {}
        self.everAppended = {}
        self.lostMessages = {}
        for consumerType in self.consumerTypes:
            self.indexPerConsumer[consumerType] = 0
            self.sent[consumerType] = 0
            self.everAppended[consumerType] = 0
            self.lostMessages[consumerType] = 0
        self.sizeSum = 0
        self.queue = []

    def hasItems(self, consumerType):
        return self.indexPerConsumer[consumerType] < len(self.queue)

    def nextMessageIndex(self, consumerType):
        index = self.indexPerConsumer[consumerType]
        foundMessage = False
        while (not foundMessage) and index < len(self.queue):
            messageFlowPattern, _, _ = self.queue[index]
            flow = Flow(messageFlowPattern)
            if (flow.isNextInFlow(consumerType, self.me)):
                foundMessage = True
            else:
                index += 1

        if (foundMessage):
            return index
        return None

    # Messages are kept in the queue until consumers of all types popped out the message.
    # An index per consumer type is maintained, to know which messages the consumer already received and conclude which message should he get now.
    def pop(self, consumerType):
        nextItemIndex = self.nextMessageIndex(consumerType)
        if (nextItemIndex is not None):
            out = self.queue[nextItemIndex]
            index = nextItemIndex + 1
            self.indexPerConsumer[consumerType] = index
            self.sent[consumerType] += 1
            anyZero = False
            for value in self.indexPerConsumer.values():
                if (value == 0):
                    anyZero = True
                    break

            if not (anyZero):
                self.queue.pop(0)
                _, _, msg = out
                self.sizeSum -= len(msg)
                for key in self.indexPerConsumer.keys():
                    self.indexPerConsumer[key] = self.indexPerConsumer[key] - 1
            return out
        return None

    def loseMessage(self):
        out = self.queue.pop(0)
        _, _, msg = out
        self.sizeSum -= len(msg)
        for key in self.indexPerConsumer.keys():
            if self.indexPerConsumer[key] > 0:
                self.indexPerConsumer[key] = self.indexPerConsumer[key] - 1
            else:
                self.lostMessages[key] += 1

    def append(self, messageFlowPattern, header, msg):
        self.sizeSum += len(msg)
        flow = Flow(messageFlowPattern)
        for consumerType in self.consumerTypes:
            if (flow.isNextInFlow(consumerType, self.me)):
                self.everAppended[consumerType] += 1
        return self.queue.append((messageFlowPattern, header, msg))

    def size(self, consumerType):
        everAppended = self.everAppended[consumerType]
        size = everAppended - self.sent[consumerType]
        return size



from hkube_python_wrapper.communication.zmq.streaming.producer.MessageQueue import MessageQueue
flow1 =   [{'source': 'B', 'next': ['C']}, {'source': 'C', 'next': ['D']}]
flow2 =  [{'source': 'B', 'next': ['A', 'C']}, {'source': 'C', 'next': ['D']},{'source': 'D', 'next': ['E']}]

def test_queue_statistics():
    queue2 = MessageQueue(['A','C'],'B')
    queue2.append(flow2,'header1','message1')
    queue2.append(flow2, 'header2', 'message2')
    queue2.append(flow2, 'header3', 'message3')
    queue2.append(flow1, 'header4', 'message4')
    assert queue2.size('A') == 3
    queue2.pop('A')
    assert queue2.size('A') == 2
    assert queue2.size('C') == 4
    assert queue2.sent['A'] == 1
    assert queue2.sent['C'] == 0
    queue2.pop('A')
    queue2.pop('A')
    queue2.pop('A')
    assert queue2.sizeSum != 0
    queue2.pop('C')
    queue2.pop('C')
    queue2.pop('C')
    queue2.pop('C')
    assert queue2.size('A') == 0
    assert queue2.size('C') == 0
    assert queue2.sent['A'] == 3
    assert queue2.sent['C'] == 4
    assert queue2.sizeSum != 0



import zmq
from .consts import consts
from hkube_python_wrapper.util.decorators import timing
from hkube_python_wrapper.util.logger import log

context = zmq.Context()


class ZMQRequest(object):
    def __init__(self, reqDetails):
        self.socket, self.connStr = self._createSocket(host=reqDetails['host'], port=reqDetails['port'])
        self.ping_socket, self.pingConnStr = self._createSocket(host=reqDetails['host'], port=int(reqDetails['port'])+1)
        self.content = reqDetails['content']
        self.timeout = int(reqDetails['timeout'])
        self.networkTimeout = int(reqDetails['networkTimeout'])

    def invokeAdapter(self):
        pong = self.ping()
        message = None
        if (pong):
            message = self.request()
        return message

    @timing
    def request(self):
        self.socket.send(self.content)
        result = self.socket.poll(self.timeout)
        if (result):
            message = self.socket.recv_multipart()
            return message
        err = 'request timed out ({timeout}) to {conn}'.format(timeout=self.timeout, conn=self.connStr)
        log.error(err)
        raise Exception(err)

    @timing
    def ping(self):
        self.ping_socket.send(consts.zmq.ping)
        result = self.ping_socket.poll(self.networkTimeout)
        if (result):
            message = self.ping_socket.recv()
            if (message == consts.zmq.pong):
                return True
        err = 'ping timed out ({timeout}) to {conn}'.format(timeout=self.networkTimeout, conn=self.pingConnStr)
        log.error(err)
        raise Exception(err)

    def close(self):
        self.socket.close()
        self.ping_socket.close()

    def _createSocket(self, host, port):
        connStr = 'tcp://' + host + ':' + str(port)
        log.info('creating socket to: {s}', s=connStr)
        socket = context.socket(zmq.REQ)
        socket.setsockopt(zmq.LINGER, 0)
        socket.connect(connStr)
        return socket, connStr



import logging
from hkube_python_wrapper.util.type_check import isString
from logging import LoggerAdapter
from hkube_python_wrapper.config import config


class Adapter(LoggerAdapter):
    def process(self, msg, kwargs):
        if isString(msg):
            return msg.format(**kwargs), {}
        return msg, kwargs


log = Adapter(logging.getLogger('wrapper'), {})


def setup():
    level = config.logging.get('level', 'INFO')
    if level in ['0', '1']:
        level = 'DEBUG'
    elif level == '2':
        level = 'INFO'
    elif level == '3':
        level = 'WARNING'
    elif level == '4':
        level = 'ERROR'
    logger = logging.getLogger('wrapper')
    logger.propagate = False
    logger.setLevel(level)
    # create console handler and set level
    ch = logging.StreamHandler()
    ch.setLevel(level)
    # create formatter
    formatter = logging.Formatter('%(name)s::%(levelname)s::%(message)s')
    # add formatter to ch
    ch.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(ch)



import time
from threading import Thread

from hkube_python_wrapper.communication.zmq.streaming.producer.ZMQProducer import ZMQProducer
from hkube_python_wrapper.communication.zmq.streaming.ZMQListener import ZMQListener
from hkube_python_wrapper.util.encoding import Encoding

PROTOCOL_TYPE_MSGPACK = 0x03
DATA_TYPE_RAW = 0x01
encoding = Encoding('msgpack')
header = encoding.createHeader(DATA_TYPE_RAW, PROTOCOL_TYPE_MSGPACK)


def test_zmq_streaming():
    def doNothing(header, msg):
        return b'5'

    count = [0, 0]
    producer = ZMQProducer(port=5559, maxMemorySize=5000, responseAcumulator=doNothing, consumerTypes=['b', 'c'], encoding=encoding, me='a')
    runThread = Thread(name="Producer", target=producer.start)
    runThread.start()

    time.sleep(1)

    def doSomething(env, header, msg):
        count[0] = count[0] + 1
        time.sleep(0.1)
        return b'5'

    def doSomething2(env, header, msg):
        count[1] = count[1] + 1
        time.sleep(0.1)
        return b'5'

    listener1 = ZMQListener('tcp://localhost:5559', doSomething, encoding, 'b')
    listener2 = ZMQListener('tcp://localhost:5559', doSomething2, encoding, 'c')
    runThread = Thread(name="Listener1", target=listener1.start)
    runThread.start()
    runThread = Thread(name="Listener2", target=listener2.start)
    runThread.start()
    time.sleep(4)
    env = [{
        "source": "a",
        "next": [
            "b", "c"
        ]
    }]
    producer.produce(header, b'bb1', env)
    producer.produce(header, b'bb2', env)
    producer.produce(header, b'bb3', env)
    producer.produce(header, b'bb4', env)
    producer.produce(header, b'bb5', env)
    time.sleep(3)

    listener1.close()
    listener2.close()
    time.sleep(1)
    producer.close()
    print(str(count[0]))
    print(str(count[1]))

    assert count[0] + count[1] == 10


if __name__ == '__main__':
    test_zmq_streaming()



import os
import random
from hkube_python_wrapper.util.encoding import Encoding

size = 1 * 1024

def test_none_encoding():
    encoding = Encoding('msgpack')
    decoded = encoding.decode(header=None, value=None)
    assert decoded is None

def test_json_encoding():
    encoding = Encoding('json')
    data = createObjectJson(size)
    encoded = encoding.encode(data, plainEncode=True)
    decoded = encoding.decode(value=encoded, plainEncode=True)
    assert data == decoded

def test_bson_encoding():
    encoding = Encoding('bson')
    data = createObject(size, size)
    (header, payload) = encoding.encode(data)
    decoded = encoding.decode(header=header, value=payload)
    assert data == decoded

def test_msgpack_encoding():
    encoding = Encoding('msgpack')
    data = create_bytearray(size)
    (header, payload) = encoding.encode(data)
    decoded = encoding.decode(header=header, value=payload)
    assert data == decoded

def test_encoding_header_payload_bytes():
    encoding = Encoding('msgpack')
    data = create_bytearray(size)
    (header, payload) = encoding.encode(data)
    decoded = encoding.decode(header=header, value=payload)
    assert data == decoded

def test_encoding_header_payload_object():
    encoding = Encoding('msgpack')
    data = createObject(size, size)
    (header, payload) = encoding.encode(data)
    decoded = encoding.decode(header=header, value=payload)
    assert data == decoded

def test_encoding_no_header_bytes():
    encoding = Encoding('msgpack')
    data = create_bytearray(size)
    (_, payload) = encoding.encode(data)
    decoded = encoding.decode(header=None, value=payload)
    assert data == decoded

def test_encoding_no_header_object():
    encoding = Encoding('msgpack')
    data = createObject(size, size)
    (_, payload) = encoding.encode(data)
    decoded = encoding.decode(header=None, value=payload)
    assert data == decoded

def test_encoding_header_in_payload_bytes():
    encoding = Encoding('msgpack')
    data = create_bytearray(size)
    (header, payload) = encoding.encode(data)
    decoded = encoding.decode(header=None, value=header + payload)
    assert data == decoded

def test_encoding_header_in_payload_object():
    encoding = Encoding('msgpack')
    data = createObject(size, size)
    (header, payload) = encoding.encode(data)
    decoded = encoding.decode(header=None, value=header + payload)
    assert data == decoded

def create_bytearray(sizeBytes):
    return b'\xdd' * (sizeBytes)

def randomString(n):
    min_lc = ord(b'a')
    len_lc = 26
    ba = bytearray(os.urandom(n))
    for i, b in enumerate(ba):
        ba[i] = min_lc + b % len_lc  # convert 0..255 to 97..122
    return ba.decode("utf-8")

def randomInt(sizeBytes):
    return random.sample(range(0, sizeBytes), sizeBytes)

def createObject(sizeBytes, sizeRandom):
    obj = {
        "bytesData": bytearray(b'\xdd' * (sizeBytes)),
        "anotherBytesData": bytearray(sizeBytes),
        "randomString": randomString(sizeRandom),
        "randomIntArray": randomInt(sizeRandom),
        "dataString": randomString(sizeRandom),
        "bool": False,
        "anotherBool": False,
        "nestedObj": {
            "dataString": randomString(sizeRandom),
            "randomIntArray": randomInt(sizeRandom)
        }
    }
    return obj

def createObjectJson(sizeRandom):
    obj = {
        "randomString": randomString(sizeRandom),
        "randomIntArray": randomInt(sizeRandom),
        "dataString": randomString(sizeRandom),
        "bool": False,
        "anotherBool": False,
        "nestedObj": {
            "dataString": randomString(sizeRandom),
            "randomIntArray": randomInt(sizeRandom)
        }
    }
    return obj



import time
import zmq
import uuid
import threading
from hkube_python_wrapper.util.logger import log

HEARTBEAT_LIVENESS = 5
HEARTBEAT_INTERVAL = 1
INTERVAL_INIT = 1
INTERVAL_MAX = 32
#  Paranoid Pirate Protocol constants
PPP_READY = b"\x01"  # Signals worker is ready
PPP_HEARTBEAT = b"\x02"  # Signals worker heartbeat
PPP_DISCONNECT = b"\x03"  # Disconnect
lock = threading.Lock()


class ZMQListener(object):

    def __init__(self, remoteAddress, onMessage, encoding, consumerType):
        self.encoding = encoding
        self.onMessage = onMessage
        self.consumerType = self.encoding.encode(consumerType, plainEncode=True)
        self.remoteAddress = remoteAddress
        self.active = True
        self.worker = None
        self.context = None
        self.heartbeat_at = None

    def worker_socket(self, remoteAddress, context=None):
        """Helper function that returns a new configured socket
           connected to the Paranoid Pirate queue"""
        if not context:
            context = zmq.Context(1)
        self.context = context
        worker = self.context.socket(zmq.DEALER)  # DEALER
        worker.setsockopt(zmq.LINGER, 0)
        identity = str(uuid.uuid4()).encode()
        worker.setsockopt(zmq.IDENTITY, identity)
        worker.setsockopt(zmq.LINGER, 0)
        worker.connect(remoteAddress)
        log.info("zmq listener connecting to {addr}", addr=remoteAddress)
        self.send(worker, [PPP_READY])
        return worker

    def send(self, workder, arr):
        arr.append(self.consumerType)
        workder.send_multipart(arr)

    def handleAMessage(self, frames):
        encodedMessageFlowPattern, header, message = frames  # pylint: disable=unbalanced-tuple-unpacking
        messageFlowPattern = self.encoding.decode(value=encodedMessageFlowPattern, plainEncode=True)
        return self.onMessage(messageFlowPattern, header, message)

    def start(self):  # pylint: disable=too-many-branches
        liveness = HEARTBEAT_LIVENESS
        interval = INTERVAL_INIT

        self.heartbeat_at = time.time() + HEARTBEAT_INTERVAL
        self.worker = self.worker_socket(self.remoteAddress)
        result = None
        while self.active:
            try:
                result = self.worker.poll(HEARTBEAT_INTERVAL * 1000)
            except Exception as e:
                if (self.active):
                    log.error('Error during poll of {addr}, {e}', addr=str(self.remoteAddress), e=str(e))
                    raise e
                break
            # Handle worker activity on backend
            if result == zmq.POLLIN:
                #  Get message
                #  - 3-part envelope + content -> request
                #  - 1-part HEARTBEAT -> heartbeat
                lock.acquire()
                try:
                    frames = self.worker.recv_multipart()
                except Exception as e:
                    lock.release()
                    if (self.active):
                        log.error('Error during receive of {addr}, {e}', addr=str(self.remoteAddress), e=str(e))
                        raise e
                    break
                if not frames:
                    lock.release()
                    if (self.active):
                        raise Exception("Connection to producer on " + self.remoteAddress + " interrupted")
                    break

                if len(frames) == 3:
                    liveness = HEARTBEAT_LIVENESS
                    try:
                        result = self.handleAMessage(frames)
                    finally:
                        lock.release()
                    try:
                        self.send(self.worker, [result])
                    except Exception as e:
                        if (self.active):
                            log.error(e)
                            raise e
                        break
                else:
                    if len(frames) == 1 and frames[0] == PPP_HEARTBEAT:
                        liveness = HEARTBEAT_LIVENESS
                        lock.release()

                    else:
                        log.error("Invalid message: {message}", message=frames)
                        liveness = HEARTBEAT_LIVENESS
                        lock.release()
                    try:
                        self.sendHeartBeat()
                    except Exception as e:
                        if (self.active):
                            log.error(e)
                        else:
                            break

                interval = INTERVAL_INIT
            else:
                liveness -= 1
                if liveness == 0:
                    log.warning("Heartbeat failure, can't reach queue of {addr}", addr=str(self.remoteAddress))
                    log.warning("Reconnecting in {interval:0.2f}", interval=interval)
                    time.sleep(interval)

                    if interval < INTERVAL_MAX:
                        interval *= 2
                    try:
                        self.worker.close()
                        self.worker = None
                    except Exception as e:
                        if (self.active):
                            log.error('Error on heartbeat failure for {addr} {e}', addr=str(self.remoteAddress), e=str(e))
                        else:
                            break
                    if (self.active):
                        self.worker = self.worker_socket(self.remoteAddress, self.context)
                    liveness = HEARTBEAT_LIVENESS

                try:
                    self.sendHeartBeat()
                except Exception as e:
                    if (self.active):
                        log.error(e)
                    else:
                        break

    def sendHeartBeat(self):
        if time.time() > self.heartbeat_at:
            self.heartbeat_at = time.time() + HEARTBEAT_INTERVAL
            self.send(self.worker, [PPP_HEARTBEAT])

    def close(self, force=True):
        # pylint: disable=too-many-nested-blocks
        if not (self.active):
            log.warning("Attempting to close inactive ZMQListener")
        else:
            self.active = False
            if (self.worker is not None):
                if not (force):
                    time.sleep(HEARTBEAT_LIVENESS + 1)
                    readAfterStopped = 0
                    try:
                        result = self.worker.poll(HEARTBEAT_INTERVAL * 1000)
                        while result == zmq.POLLIN:
                            lock.acquire()
                            try:
                                frames = self.worker.recv_multipart()
                                if len(frames) == 3:
                                    self.handleAMessage(frames)
                                    readAfterStopped += 1
                                    log.warning('Read after stop {readAfterStopped}', readAfterStopped=readAfterStopped)
                                    self.send(self.worker, [PPP_DISCONNECT])
                            finally:
                                lock.release()
                            result = self.worker.poll(HEARTBEAT_INTERVAL * 1000)
                    except Exception as e:
                        log.error('Error on zmqListener close {e}', e=str(e))
                    else:
                        try:
                            self.send(self.worker, [PPP_DISCONNECT])
                        except Exception as e:
                            log.error('Error on sending disconnect {e}', e=str(e))
                self.worker.close()
                self.context.destroy()



import time

HEARTBEAT_LIVENESS = 3  # 3..5 is reasonable
HEARTBEAT_INTERVAL = 1.0  # Seconds


class Worker(object):
    def __init__(self, address):
        self.address = address
        self.expiry = time.time() + HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS



from hkube_python_wrapper.communication.zmq.streaming.ZMQListener import ZMQListener
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.util.DaemonThread import DaemonThread
from hkube_python_wrapper.util.logger import log
import time


class MessageListener(DaemonThread):

    def __init__(self, options, receiverNode, errorHandler=None):
        self.errorHandler = errorHandler
        remoteAddress = options['remoteAddress']
        encodingType = options['encoding']
        self._encoding = Encoding(encodingType)
        self.adapater = ZMQListener(remoteAddress, self.onMessage, self._encoding, receiverNode)
        self.messageOriginNodeName = options['messageOriginNodeName']

        self.messageListeners = []
        DaemonThread.__init__(self, "MessageListener-" + str(self.messageOriginNodeName))

    def registerMessageListener(self, listener):
        self.messageListeners.append(listener)

    def onMessage(self, messageFlowPattern, header, msg):
        start = time.time()
        decodedMsg = self._encoding.decode(header=header, value=msg)
        for listener in self.messageListeners:
            try:
                listener(messageFlowPattern, decodedMsg, self.messageOriginNodeName)
            except Exception as e:
                log.error('Error during MessageListener onMessage {e}', e=str(e))

        end = time.time()
        duration = float((end - start) * 1000)
        return self._encoding.encode({'duration': round(duration, 4)}, plainEncode=True)

    def run(self):
        log.info("Start receiving from {node}", node=self.messageOriginNodeName)
        try:
            self.adapater.start()
        except Exception as e:
            if (self.errorHandler):
                self.errorHandler.sendError(e)

    def close(self, force=True):
        self.adapater.close(force)
        self.messageListeners = []



from hkube_python_wrapper.cache.caching import Cache


def test_reaching_limit():
    cache = Cache({"maxCacheSize": 4})
    value1 = {"bytes": bytearray(1000000)}
    cache.update("task1", value1, 1000000)
    assert len(cache._cache) == 1
    cache.update("task2", value1, 1000000)
    cache.update("task3", value1, 1000000)
    assert len(cache._cache) == 3
    cache.update("task4", value1, 1500000)
    assert len(cache._cache) == 3
    assert "task4" in cache
    assert "task1" not in cache


def test_too_large_message():
    cache = Cache({"maxCacheSize": 5})
    value1 = {"bytes": bytearray(1000000)}
    value2 = {"bytes": bytearray(5000000)}
    result = cache.update("task1", value1, 1000000)
    assert result == True
    assert len(cache._cache) == 1
    result = cache.update("task2", value2, 6000000)
    assert result == False
    assert len(cache._cache) == 1


def test_get_all():
    cache = Cache({"maxCacheSize": 5})
    value1 = {"data": "1"}
    value2 = {"data": "2"}
    result = cache.update("task1", value1)
    assert result == True
    assert len(cache._cache) == 1
    cache.update("task2", value2)
    assert len(cache._cache) == 2
    tasksNotInCache, valuesInCache = cache.getAll(['task1', 'task2', 'task3', 'task4'])
    assert len(tasksNotInCache) == 2
    assert "task3" in set(tasksNotInCache)
    assert "task4" in set(tasksNotInCache)
    assert len(valuesInCache) == 2
    data = list(map(lambda value: value.get('data'), valuesInCache))
    assert "1" in set(data)
    assert "2" in set(data)


if __name__ == '__main__':
    test_get_all()



class FifoArray(object):
    def __init__(self, size):
        self.arr = []
        self.size = size

    def append(self, obj):
        self.arr.append(obj)
        if (len(self.arr) > self.size):
            self.arr.pop(0)

    def getAsArray(self):
        return self.arr

    def reset(self):
        self.arr = []



from hkube_python_wrapper.communication.streaming.StreamingManager import StreamingManager
from hkube_python_wrapper.communication.streaming.MessageListener import MessageListener
from hkube_python_wrapper.communication.streaming.MessageProducer import MessageProducer
import time

parsedFlows = {'analyze': [{'source': 'A', 'next': ['B']}, {'source': 'B', 'next': ['C']}, {'source': 'C', 'next': ['D']}], 'master': [{'source': 'B', 'next': ['A', 'C']}, {'source': 'C', 'next': ['D']}, {'source': 'D', 'next': ['E']}]}
parents = [{'nodeName': 'A', 'address': {'host': '127.0.0.1', 'port': '9326'}, 'type': 'Add'}]


def test_streaming_manager():
    producer_configA = {'port': 9326, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
    producer_configB = {'port': 9426, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
    listen_toB_config = {'remoteAddress': 'tcp://localhost:9426', 'encoding': 'msgpack', 'messageOriginNodeName': 'B'}
    listen_config = {'encoding': 'msgpack'}
    parents = [{'nodeName': 'A', 'address': {'host': '127.0.0.1', 'port': '9326'}, 'type': 'Add'}]
    streamingManagaerA = StreamingManager(None)
    streamingManagaerA.setParsedFlows(parsedFlows, 'analyze')

    streamingManagaerB = StreamingManager(None)
    streamingManagaerB.setParsedFlows(parsedFlows, 'analyze')

    messageListener = MessageListener(listen_toB_config, receiverNode='C')
    resultsAtC = {}

    def onMessageB2C(flow, msg, origin):
        resultsAtC['flowLength'] = len(flow)
        resultsAtC['flowFirstSource'] = flow[0]['source']
        resultsAtC['msg'] = msg
        print("atOnmessageC")

    def onMessageAtB(msg, origin):
        print('atOnmessageB')
        streamingManagaerB.sendMessage('stam_klum')

    def statsInvoked(args):
        print('stats')

    try:
        streamingManagaerA.setupStreamingProducer(statsInvoked, producer_configA, ['B'], 'A')
        streamingManagaerB.setupStreamingProducer(statsInvoked, producer_configB, ['C'], 'B')
        streamingManagaerB.setupStreamingListeners(listen_config, parents, 'B')
        streamingManagaerB.registerInputListener(onMessageAtB)
        time.sleep(1)
        messageListener.registerMessageListener(onMessageB2C)
        streamingManagaerB.startMessageListening()

        messageListener.start()
        time.sleep(1)
        streamingManagaerA.sendMessage('klum')
        time.sleep(4)
        assert resultsAtC['flowLength'] == 1
        assert resultsAtC['flowFirstSource'] == 'C'
        assert resultsAtC['msg'] == 'stam_klum'
        # send to none default flow
        streamingManagaerB.sendMessage('klum', 'master')
        time.sleep(1)
        assert resultsAtC['flowLength'] == 2
        assert resultsAtC['flowFirstSource'] == 'C'
    finally:
        streamingManagaerA.stopStreaming()
        streamingManagaerB.stopStreaming()
        messageListener.close()


def test_Messaging():
    try:
        producer_config = {'port': 9536, 'messagesMemoryBuff': 5000, 'encoding': 'msgpack', 'statisticsInterval': 1}
        listenr_config = {'remoteAddress': 'tcp://localhost:9536', 'encoding': 'msgpack', 'messageOriginNodeName': 'b'}
        asserts = {}
        asserts['responses'] = 0
        messageProducer = MessageProducer(producer_config, ['a'], 'b')

        def onStatistics(statistics):
            asserts['stats'] = statistics
            asserts['responses'] = int(statistics[0]['responses'])

        messageProducer.registerStatisticsListener(onStatistics)
        time.sleep(3)

        def onMessage(envelope, msg, origin):
            # pylint: disable=unused-argument
            if (type(msg) == type(dict())):
                asserts['field1'] = msg['field1']
            time.sleep(0.1)

        messageProducer.start()
        env = [{
            "source": "b",
            "next": [
                "a"
            ]
        }, {
            "source": "a",
            "next": [
                "c"
            ]
        }]
        messageProducer.produce(env, {'field1': 'value1'})
        messageProducer.produce(env, {'field1': 'value1'})
        messageProducer.produce(env, {'field1': 'value1'})
        time.sleep(2)
        assert asserts['stats'][0]['queueSize'] == 3
        assert asserts['stats'][0]['sent'] == 0
        messageListener = MessageListener(listenr_config, receiverNode='a')
        messageListener.registerMessageListener(onMessage)
        messageListener.start()
        for _ in range(1,5):
            if ( asserts.get('field1')):
                break
            time.sleep(1)
        time.sleep(4)
        assert asserts['field1'] == 'value1'
        assert asserts['stats'][0]['queueSize'] == 0
        assert asserts['stats'][0]['sent'] == 3
        assert asserts['responses'] == 3
        print('done\n')

        def getHello(envelope, msg, origin):
            if (msg == b'Hello'):
                asserts['gotHello'] = True
                asserts['envelope'] = envelope

        messageListener.registerMessageListener(getHello)
        messageProducer.produce(env, b'Hello')
        time.sleep(3)
        assert asserts['gotHello']
        assert len(asserts['envelope']) == 1
        assert asserts['envelope'][0]['source'] == 'a'
    finally:
        messageProducer.close()
        messageListener.close()




import time
import types

from hkube_python_wrapper.storage.storage_manager import StorageManager
from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper import Algorunner
from tests.configs import config
from hkube_python_wrapper.tracing.tracer import Tracer

algorithmName = 'eval-alg'
subpipelineName = 'simple'

encoding = Encoding(config.storage['encoding'])
inp1 =  {'storageInfo': {'path': 'c'}}
inp2 = {'storageInfo': {'path': 'a'}}
storageMock = {'a': [{'nodeName': 'node1','info':{'path':'b','isBigData':True}}],'b':'stam','c': [{'nodeName': 'node1','result':'originalRelsult'}]}
outp2 =  [{'nodeName': 'node1','info':{'path':'b','isBigData':True},'result':'stam'}]

def start(args, hkubeApi=None):
    # print('start called')
    waiter1 = hkubeApi.start_algorithm(algorithmName, inp1)
    waiter2 = hkubeApi.start_stored_subpipeline(subpipelineName, inp2)
    res = [waiter1, waiter2]
    return res


def storageManagerMocke(algorunner, options):
    if (algorunner._storage != 'v1'):
        algorunner._initDataServer(options)
    algorunner._initDataAdapter(options)
    algorunner._dataAdapter.tryGetDataFromPeerOrStorage = lambda response: storageMock.get(response.get('storageInfo').get('path'))


def test_callback():
    sm = StorageManager(config.storage)
    algorunner = Algorunner()

    algorunner._initStorage = types.MethodType(storageManagerMocke, algorunner)
    algorunner.loadAlgorithmCallbacks(start, options=config)
    algorunner.connectToWorker(config)
    time.sleep(3)
    (header, payload) = sm.storage.get({"path": "local-hkube/jobId/taskId"})
    decoded = encoding.decode(header=header, value=payload)
    assert decoded[0] == storageMock.get('c')
    assert decoded[1] == outp2
    time.sleep(2)
    Tracer.instance.close()
    algorunner.close()



#
# Paranoid Pirate queue
#
#   Author: Daniel Lundin <dln(at)eintr(dot)org>
#
import time
from .Flow import Flow
from .Worker import Worker
from .WorkerQueue import WorkerQueue
from .MessageQueue import MessageQueue
from hkube_python_wrapper.util.logger import log
import zmq

HEARTBEAT_LIVENESS = 5  # 3..5 is reasonable
HEARTBEAT_INTERVAL = 1.0  # Seconds

#  Paranoid Pirate Protocol constants
PPP_READY = b"\x01"  # Signals worker is ready
PPP_HEARTBEAT = b"\x02"  # Signals worker heartbeat
PPP_DISCONNECT = b"\x03"  # Disconnect


class ZMQProducer(object):
    def __init__(self, port, maxMemorySize, responseAcumulator, consumerTypes, encoding, me):
        self.me = me
        self.encoding = encoding
        self.responseAcumulator = responseAcumulator
        self.maxMemorySize = maxMemorySize
        self.port = port
        self.consumerTypes = consumerTypes
        self.messageQueue = MessageQueue(consumerTypes, self.me)
        context = zmq.Context(1)
        self._backend = context.socket(zmq.ROUTER)  # ROUTER
        self._backend.bind("tcp://*:" + str(port))  # For workers
        log.info("Producer listening on tcp://*:{port}", port=port)
        self.active = True
        self.watingForResponse = []

    def produce(self, header, message, messageFlowPattern=[]):
        while (self.messageQueue.sizeSum > self.maxMemorySize):
            self.messageQueue.loseMessage()
        self.messageQueue.append(messageFlowPattern, header, message)

    def start(self):  # pylint: disable=too-many-branches
        poll_workers = zmq.Poller()
        poll_workers.register(self._backend, zmq.POLLIN)
        workers = WorkerQueue(self.consumerTypes)

        heartbeat_at = time.time() + HEARTBEAT_INTERVAL
        while self.active:
            poller = poll_workers
            try:
                socks = dict(poller.poll(HEARTBEAT_INTERVAL * 1000))
            except Exception as e:
                if (self.active):
                    log.error(e)
                else:
                    break
            # Handle worker activity on self._backend
            if socks.get(self._backend) == zmq.POLLIN:
                # Use worker address for LRU routing
                try:
                    frames = self._backend.recv_multipart()
                except Exception as e:
                    if (self.active):
                        log.error(e)
                    else:
                        break
                if not frames:
                    if (self.active):
                        raise Exception("Unexpected router no frames on receive, no address frame")
                    break
                address = frames[0]
                consumerType = self.encoding.decode(value=frames[2], plainEncode=True)
                if not consumerType in self.consumerTypes:
                    log.warning("Producer got message from unknown consumer: {consumerType}, dropping the message", consumerType=consumerType)
                    continue
                if frames[1] not in (PPP_READY, PPP_HEARTBEAT, PPP_DISCONNECT):
                    self.watingForResponse.remove(address)
                    self.responseAcumulator(frames[1], consumerType)
                if not address in self.watingForResponse:
                    workers.ready(Worker(address), consumerType)
                else:
                    if frames[1] == PPP_DISCONNECT:
                        self.watingForResponse.remove(address)

                # Validate control message, or return reply to client
                msg = frames[1:]
                if time.time() >= heartbeat_at:
                    for consumerType, workersOfType in workers.queues.items():
                        for worker in workersOfType:
                            msg = [worker, PPP_HEARTBEAT]
                            try:
                                self._backend.send_multipart(msg)
                            except Exception as e:
                                if (self.active):
                                    log.error(e)
                                else:
                                    break
                    heartbeat_at = time.time() + HEARTBEAT_INTERVAL
            for consumerType, workerQueue in workers.queues.items():
                if (workerQueue):
                    poped = self.messageQueue.pop(consumerType)
                    if (poped):
                        messageFlowPattern, header, payload = poped
                        flow = Flow(messageFlowPattern)
                        frames = [self.encoding.encode(flow.getRestOfFlow(self.me), plainEncode=True), header, payload]
                        identity = workers.next(consumerType)
                        self.watingForResponse.append(identity)
                        frames.insert(0, identity)
                        try:
                            self._backend.send_multipart(frames)
                        except Exception as e:
                            if (self.active):
                                log.error(e)
                            else:
                                break
            workers.purge()

    def queueSize(self, consumerSize):
        return self.messageQueue.size(consumerSize)

    def sent(self, consumerType):
        return self.messageQueue.sent[consumerType]

    def close(self, force=True):
        stillInQueue = 0
        while self.messageQueue.queue and not force:
            stillInQueue += 1
            time.sleep(1)
        log.info('Closing dealt with {stillInQueue} more', stillInQueue=stillInQueue)
        self.active = False
        time.sleep(HEARTBEAT_LIVENESS + 1)
        if not force:
            self._backend.close(10)
        else:
            self._backend.close(0)



class Flow:
    def __init__(self, flow):
        self.flow = flow

    def isNextInFlow(self, next, currentName):
        current = self._getCurrent(currentName)
        if (current is None):
            return False
        return next in current['next']

    def getRestOfFlow(self, currentName):
        current = self._getCurrent(currentName)
        if (current is None):
            return []
        flowcopy = self.flow[:]
        flowcopy.remove(current)
        return flowcopy

    def _getCurrent(self, currentName):
        for node in self.flow:
            if node['source'] == currentName:
                current = node
                return current
        return None



from __future__ import print_function, division, absolute_import
import time
from hkube_python_wrapper.util.object_path import getPath
import hkube_python_wrapper.util.type_check as typeCheck
from hkube_python_wrapper.wrapper.messages import messages
from .execution import Execution
from .waitFor import WaitForData
from hkube_python_wrapper.util.queueImpl import Empty
from hkube_python_wrapper.util.logger import log


class HKubeApi:
    """Hkube interface for code-api operations"""

    def __init__(self, wc, wrapper, dataAdapter, storage, streamingManager):
        self._wc = wc
        self._wrapper = wrapper
        self._dataAdapter = dataAdapter
        self._storage = storage
        self._executions = {}
        self._lastExecId = 0
        self.streamingManager = streamingManager

    def registerInputListener(self, onMessage):
        self.streamingManager.registerInputListener(onMessage)

    def startMessageListening(self):
        self.streamingManager.startMessageListening()

    def sendMessage(self, msg, flowName=None):
        self.streamingManager.sendMessage(msg, flowName)

    def stopStreaming(self, force=True):
        self.streamingManager.stopStreaming(force)

    def isListeningToMessages(self):
        return self.streamingManager.listeningToMessages

    def _generateExecId(self):
        self._lastExecId += 1
        return str(self._lastExecId)

    def algorithmExecutionDone(self, data):
        execId = data.get('execId')
        self._handleExecutionDone(execId, data)

    def subPipelineDone(self, data):
        subPipelineId = data.get('subPipelineId')
        self._handleExecutionDone(subPipelineId, data)

    def dataSourceResponse(self, data):
        requestId = data.get('requestId')
        execution = self._executions.get(requestId)
        try:
            error = data.get('error')
            dataSource = None
            if (not error):
                dataSource = data.get('response')
            execution.waiter.set((error, dataSource))
        except Exception as e:
            execution.waiter.set(e)
        finally:
            self._executions.pop(requestId)

    def _handleExecutionDone(self, execId, data):
        execution = self._executions.get(execId)
        # pylint: disable=too-many-nested-blocks
        try:
            error = data.get('error')
            if (error):
                execution.waiter.set(error)

            elif (execution.includeResult):
                response = data.get('response')
                result = response
                if (typeCheck.isDict(response) and response.get('storageInfo') and self._storage == 'v3'):
                    result = self._dataAdapter.tryGetDataFromPeerOrStorage(
                        response)
                if self._storage == 'v3' and typeCheck.isList(result):
                    for node in result:
                        if typeCheck.isDict(node) and getPath(node, 'info.isBigData') is True:
                            nodeResult = self._dataAdapter.tryGetDataFromPeerOrStorage({"storageInfo": node['info']})
                            node['result'] = nodeResult
                execution.waiter.set(result)
            else:
                execution.waiter.set(None)

        except Exception as e:
            execution.waiter.set(e)
        finally:
            self._executions.pop(execId)

    def start_algorithm(self, algorithmName, input=[], includeResult=True):
        """Starts algorithm execution.

        starts an invocation of algorithm with input, and waits for results

        Args:
            algorithmName (string): The name of the algorithm to start.
            input (array): Optional input for the algorithm.
            includeResult (bool): if True, returns the result of the algorithm execution.
        Returns:
            Returns the result of the algorithm
        """
        log.info('start_algorithm called with {name}', name=algorithmName)
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startAlgorithmExecution,
            "data": {
                "execId": execId,
                "algorithmName": algorithmName,
                "input": input,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)

        return self._waitForResult(execution)

    def getDataSource(self, dataSource):
        log.info('getDataSource called')
        requestId = self._generateExecId()
        execution = Execution(requestId, False, WaitForData(True))
        self._executions[requestId] = execution

        message = {
            "command": messages.outgoing.dataSourceRequest,
            "data": {
                "requestId": requestId,
                "dataSource": dataSource
            }
        }
        self._wc.send(message)

        return self._waitForResult(execution)

    def start_stored_subpipeline(self, name, flowInput={}, includeResult=True):
        """Starts pipeline execution.

        starts an invocation of a sub-pipeline with input, and waits for results

        Args:
            name (string): The name of the pipeline to start.
            flowInput (dict): Optional flowInput for the pipeline.
            includeResult (bool): if True, returns the result of the pipeline execution.
                default: True
        Returns:
            Returns the result of the pipeline
        """
        log.info('start_stored_subpipeline called with {name}', name=name)
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startStoredSubPipeline,
            "data": {
                "subPipeline": {
                    "name": name,
                    "flowInput": flowInput
                },
                "subPipelineId": execId,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)
        return self._waitForResult(execution)

    def start_raw_subpipeline(self, name, nodes, flowInput, options={}, webhooks={}, includeResult=True):
        """Starts pipeline execution.

        starts an invocation of a sub-pipeline with input, nodes, options, and optionally waits for results

        Args:
            name (string): The name of the pipeline to start.
            nodes (array): array of node definitions (like in the pipeline descriptor)
            options (dict): pipeline options (like in the pipeline descriptor)
            webhooks (dict): webhook options (like in the pipeline descriptor)
            flowInput (dict): flowInput for the pipeline.
            includeResult (bool): if True, returns the result of the pipeline execution.
                default: True
        Returns:
            Returns the result of the pipeline
        """
        log.info('start_raw_subpipeline called with {name}', name=name)
        execId = self._generateExecId()
        execution = Execution(execId, includeResult, WaitForData(True))
        self._executions[execId] = execution

        message = {
            "command": messages.outgoing.startRawSubPipeline,
            "data": {
                "subPipeline": {
                    "name": name,
                    "nodes": nodes,
                    "options": options,
                    "webhooks": webhooks,
                    "flowInput": flowInput
                },
                "subPipelineId": execId,
                "includeResult": includeResult
            }
        }
        self._wc.send(message)
        return self._waitForResult(execution)

    def _waitForResult(self, execution):
        while not execution.waiter.ready():
            try:
                (command, data) = self._wrapper.get_message(False)
                self._wrapper.handle(command, data)
            except Empty:
                pass
            time.sleep(0.01)
        return execution.waiter.get()



import time
from threading import Thread

from hkube_python_wrapper.util.encoding import Encoding
from hkube_python_wrapper.communication.zmq.streaming.producer.ZMQProducer import ZMQProducer
from hkube_python_wrapper.util.fifo_array import FifoArray
from hkube_python_wrapper.util.DaemonThread import DaemonThread
from hkube_python_wrapper.util.logger import log

RESPONSE_CACHE = 2000


class MessageProducer(DaemonThread):
    def __init__(self, options, consumerNodes, me):
        self.nodeNames = consumerNodes
        port = options['port']
        maxMemorySize = options['messagesMemoryBuff'] * 1024 * 1024
        encodingType = options['encoding']
        statisticsInterval = options['statisticsInterval']
        self._encoding = Encoding(encodingType)
        self.adapter = ZMQProducer(port, maxMemorySize, self.responseAccumulator, consumerTypes=self.nodeNames, encoding=self._encoding, me=me)
        self.responsesCache = {}
        self.responseCount = {}
        self.active = True
        self.printStatistics = 0
        for nodeName in consumerNodes:
            self.responsesCache[nodeName] = FifoArray(RESPONSE_CACHE)
            self.responseCount[nodeName] = 0
        self.listeners = []

        def sendStatisticsEvery(interval):
            while (self.active):
                self.sendStatistics()
                time.sleep(interval)

        if (self.nodeNames):
            runThread = Thread(name="Statistics", target=sendStatisticsEvery, args=[statisticsInterval])
            runThread.daemon = True
            runThread.start()
        DaemonThread.__init__(self, "MessageProducer")

    def produce(self, meesageFlowPattern, obj):
        header, encodedMessage = self._encoding.encode(obj)
        self.adapter.produce(header, encodedMessage, messageFlowPattern=meesageFlowPattern)

    def responseAccumulator(self, response, consumerType):
        decodedResponse = self._encoding.decode(value=response, plainEncode=True)
        self.responseCount[consumerType] += 1
        duration = decodedResponse['duration']
        self.responsesCache[consumerType].append(float(duration))

    def resetResponseCache(self, consumerType):
        responsePerNode = self.responsesCache[consumerType].getAsArray()
        self.responsesCache[consumerType].reset()
        return responsePerNode

    def getResponseCount(self, consumerType):
        return self.responseCount[consumerType]

    def registerStatisticsListener(self, listener):
        self.listeners.append(listener)

    def sendStatistics(self):
        statistics = []
        for nodeName in self.nodeNames:
            queueSize = self.adapter.queueSize(nodeName)
            sent = self.adapter.sent(nodeName)
            singleNodeStatistics = {"nodeName": nodeName, "sent": sent, "queueSize": queueSize,
                                    "durations": self.resetResponseCache(nodeName),
                                    "responses": self.getResponseCount(nodeName),
                                    "dropped": self.adapter.messageQueue.lostMessages[nodeName]}
            statistics.append(singleNodeStatistics)
        for listener in self.listeners:
            listener(statistics)
        for singleNodeStatisticsForPrint in statistics:
            singleNodeStatisticsForPrint['durations'] = singleNodeStatisticsForPrint['durations'][:10]
        if (self.printStatistics % 10 == 0):
            log.debug("statistics {stats}", stats=str(statistics))
        self.printStatistics += 1

    def run(self):
        self.adapter.start()

    def close(self, force=True):
        if not (self.active):
            log.warning("Attempting to close inactive MessageProducer")
        else:
            self.adapter.close(force)
            self.active = False



import time
import threading

class statelessAlgoWrapper(object):
    def __init__(self, algo):
        self._hkubeApi = None
        self.originalAlgorithm = algo
        self.options = None
        self.active = False
        self.error = None
        self.threadLocalStorage = threading.local()

    def _invokeAlgorithm(self, msg, origin):
        if not (self.originalAlgorithm.get('init') is None):
            self.originalAlgorithm['init'](msg)
            # TODO should init be called upon every message
        options = {}
        options.update(self.options)
        options['streamInput'] = {'message': msg, 'origin': origin}
        try:
            result = self.originalAlgorithm['start'](options, self._hkubeApi)
            if (self.options['childs']):
                self._hkubeApi.sendMessage(result)
        except Exception as e:
            self.error = e

    def start(self, options, hkube_api):
        # pylint: disable=unused-argument
        self._hkubeApi = hkube_api
        self._hkubeApi.registerInputListener(onMessage=self._invokeAlgorithm)
        self._hkubeApi.startMessageListening()
        self.active = True
        self.error = None
        while (self.active):
            if (self.error is not None):
                raise self.error  # pylint: disable=raising-bad-type
            time.sleep(1)

    def init(self, options):
        self.options = options

    def exit(self, data):
        self.active = False
        if not (self.originalAlgorithm.get('exit') is None):
            self.originalAlgorithm['exit'](data)

    def stop(self, data):
        self.active = False
        if self.originalAlgorithm.get('stop'):
            self.originalAlgorithm['stop'](data)



from __future__ import print_function, division, absolute_import
import logging
from hkube_python_wrapper.codeApi.hkube_api import HKubeApi
from hkube_python_wrapper.wrapper.algorunner import Algorunner
from hkube_python_wrapper.util.logger import setup

name = "hkube_python_wrapper"
setup()



from collections import OrderedDict
import time
from hkube_python_wrapper.util.logger import log


class WorkerQueue(object):
    def __init__(self, consumerTypes):
        self.queues = {}
        for consumerType in consumerTypes:
            self.queues[consumerType] = OrderedDict()

    def ready(self, worker, consumerType):
        self.queues[consumerType].pop(worker.address, None)
        self.queues[consumerType][worker.address] = worker

    def purge(self):
        """Look for & kill expired workers."""
        t = time.time()
        expired = []
        for type, queue in self.queues.items():
            for address, worker in queue.items():
                if t > worker.expiry:  # Worker expired
                    expired.append((address, type))
            for (address, consumerType) in expired:
                log.warning("Idle worker expired: {address}", address=address)
                self.queues[consumerType].pop(address, None)

    def next(self, type):
        address, worker = self.queues[type].popitem(False)  # pylint: disable=unused-variable
        return address



import threading

from .MessageListener import MessageListener
from .MessageProducer import MessageProducer
from hkube_python_wrapper.util.logger import log


class StreamingManager():
    threadLocalStorage = threading.local()

    def __init__(self, errorHandler):
        self.errorHandler = errorHandler
        self.messageProducer = None
        self._messageListeners = dict()
        self._inputListener = []
        self.listeningToMessages = False
        self.parsedFlows = {}
        self.defaultFlow = None
        self.listenerLock = threading.Lock()

    def setParsedFlows(self, flows, defaultFlow):
        self.parsedFlows = flows
        self.defaultFlow = defaultFlow

    def sendError(self, e):
        self.errorHandler.sendError(e)

    def setupStreamingProducer(self, onStatistics, producerConfig, nextNodes, me):
        self.messageProducer = MessageProducer(producerConfig, nextNodes, me)
        self.messageProducer.registerStatisticsListener(onStatistics)
        if (nextNodes):
            self.messageProducer.start()

    def setupStreamingListeners(self, listenerConfig, parents, nodeName):
        self.listenerLock.acquire()
        try:
            log.debug("parents {parents}", parents=str(parents))
            for predecessor in parents:
                remoteAddress = predecessor['address']
                remoteAddressUrl = 'tcp://{host}:{port}'.format(host=remoteAddress['host'], port=remoteAddress['port'])
                if (predecessor['type'] == 'Add'):
                    options = {}
                    options.update(listenerConfig)
                    options['remoteAddress'] = remoteAddressUrl
                    options['messageOriginNodeName'] = predecessor['nodeName']
                    listener = MessageListener(options, nodeName, self)
                    listener.registerMessageListener(self._onMessage)
                    self._messageListeners[remoteAddressUrl] = listener
                    if (self.listeningToMessages):
                        listener.start()
                if (predecessor['type'] == 'Del'):
                    if (self.listeningToMessages):
                        try:
                            self._messageListeners[remoteAddressUrl].close()
                        except Exception as e:
                            log.error('another Exception: {e}', e=str(e))
                        del self._messageListeners[remoteAddressUrl]
        finally:
            self.listenerLock.release()

    def registerInputListener(self, onMessage):
        self._inputListener.append(onMessage)

    def _onMessage(self, messageFlowPattern, msg, origin):
        self.threadLocalStorage.messageFlowPattern = messageFlowPattern
        for listener in self._inputListener:
            try:
                listener(msg, origin)
            except Exception as e:
                log.error("hkube_api message listener through exception: {e}", e=str(e))
        self.threadLocalStorage.messageFlowPattern = []

    def startMessageListening(self):
        self.listeningToMessages = True
        self.listenerLock.acquire()
        try:
            for listener in self._messageListeners.values():
                if not (listener.is_alive()):
                    listener.start()
        finally:
            self.listenerLock.release()

    def sendMessage(self, msg, flowName=None):
        if (self.messageProducer is None):
            raise Exception('Trying to send a message from a none stream pipeline or after close had been applied on algorithm')
        if (self.messageProducer.nodeNames):
            parsedFlow = None
            if (flowName is None):
                if hasattr(self.threadLocalStorage, 'messageFlowPattern') and self.threadLocalStorage.messageFlowPattern:
                    parsedFlow = self.threadLocalStorage.messageFlowPattern
                else:
                    if (self.defaultFlow is None):
                        raise Exception("Streaming default flow is None")
                    flowName = self.defaultFlow
            if not (parsedFlow):
                parsedFlow = self.parsedFlows.get(flowName)
            if (parsedFlow is None):
                raise Exception("No such flow " + flowName)
            self.messageProducer.produce(parsedFlow, msg)

    def stopStreaming(self, force=True):
        if (self.listeningToMessages):
            self.listenerLock.acquire()
            try:
                for listener in self._messageListeners.values():
                    listener.close(force)
                self._messageListeners = dict()
            finally:
                self.listeningToMessages = False
                self.listenerLock.release()
        self._inputListener = []
        if (self.messageProducer is not None):
            self.messageProducer.close(force)
            self.messageProducer = None

    def clearMessageListeners(self):
        self.listenerLock.acquire()
        self._messageListeners = dict()
        self.listenerLock.release()
